import os


class Workspace:
    '''
    Docker workspaces are unique in our world, as they can be one of two
    context dependent things: A Docker build directory, containing only a
    single Dockerfile, or they can be part of a formation using docker-compose
    in which they warehouse a docker-compose.yml file.

    Under most situations we only care about the context the charm author
    wishes to be in, and what implications that has on the workspace to be
    valid.

    This method simply exposes an overrideable object to determine these
    characteristics.
    '''
    def __init__(self, path, context="compose"):
        self.path = path
        self.context = context

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.path

    def validate(self):
        dcyml = os.path.isfile("{}/docker-compose.yml".format(self.path))
        dcyaml = os.path.isfile("{}/docker-compose.yaml".format(self.path))
        dfile = os.path.isfile("{}/Dockerfile".format(self.path))

        if self.context == "compose":
            if not dcyml and not dcyaml:
                msg = "Missing yaml definition: docker-compose.yml"
                raise OSError(msg)
        else:
            if not dfile:
                msg = "Missing Dockerfile"
                raise OSError(msg)
        return True



import os
import subprocess

from shlex import split

from .workspace import Workspace


class Docker:
    '''
    Wrapper class to communicate with the Docker daemon on behalf of
    a charmer. Provides stateless operations of a running docker daemon
    '''

    def __init__(self, socket="unix:///var/run/docker.sock", workspace=None):
        '''
        :param socket: URI to the Docker daemon socket
            default: unix:///var/run/docker.sock

        :param workspace: Path to directory containing a Dockerfile
            default: None
        '''
        self.socket = socket
        if workspace:
            self.workspace = Workspace(workspace)

    def running(self):
        '''
        Predicate method to determine if the daemon we are talking to is
        actually online and recieving events.

        ex: bootstrap = Docker(socket="unix:///var/run/docker-boostrap.sock")
        bootstrap.running()
        > True
        '''
        # TODO: Add TCP:// support for running check
        return os.path.isfile(self.socket)

    def run(self, image, options=[], commands=[], arg=[]):
        '''
        Docker Run exposed as a method. This wont be as natural as the
        command line docker experience.

        Docker CLI output example:
        Usage:	docker run [OPTIONS] IMAGE [COMMAND] [ARG...]

        :param image: string of the container to pull from the registry,
                        eg: ubuntu:latest
        :param options:  array of string  options, eg: ['-d', '-v /tmp:/tmp']
        :param commands:  array of string commands, eg: ['ls']
        :param arg:  array of string command args, eg: ['-al']
        '''
        options = ' '.join(options)
        command = ' '.join(commands)
        args = ' '.join(arg)
        cmd = "docker run {0} {1} {2} {3}".format(
            options, image, command, args)

        try:
            subprocess.check_output(split(cmd))
        except subprocess.CalledProcessError as expect:
            print("Error: ", expect.returncode, expect.output)

    def login(self, user, password, email):
        '''
        Docker login exposed as a method.

        :param user:  Username in the registry
        :param password: - Password for the registry
        :param email: - Email address on account (dockerhub)
        '''
        cmd = ['docker', 'login', '-u', user, '-p', password, '-e', email]
        subprocess.check_call(cmd)

    def logs(self, container_id, raise_on_failure=False):
        '''
        Docker logs exposed as a method.

        :param container_id: - UUID for the container to fetch logs
        '''
        cmd = ['docker', 'logs', container_id]
        output = subprocess.check_output(cmd)

        return output.decode('ascii', 'ignore')

    def ps(self):
        '''
        return a string of docker status output
        '''
        cmd = ['docker', 'ps']
        return subprocess.check_output(cmd)

    def pull(self, image):
        '''
        Pull an image from the docker hub
        '''
        cmd = ['docker', 'pull', image]
        return subprocess.check_output(cmd)



from contextlib import contextmanager
from shlex import split
from subprocess import check_output
import os


def run(cmd, workspace):
    '''
    wrapper for executing the commands generated by the class members.
    commands are passed through shlex.parse for convenience.

    :param cmd: - String of the command to run. eg: echo "hello world".

    :returns: STDOUT of command execution

    :usage: c.run('docker-compose ps')
    '''
    with chdir("{}".format(workspace)):
        out = check_output(split(cmd))
        return out


# This is helpful for setting working directory context
@contextmanager
def chdir(path):
    '''Change the current working directory to a different directory to run
    commands and return to the previous directory after the command is done.'''
    old_dir = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(old_dir)



from charmhelpers.core import unitdata


class DockerOpts:
    '''
    DockerOptsManager - A Python class for managing the DEFAULT docker
    options on a daemon dynamically. As a docker daemon integrates with more
    services it becomes quickly unweidly to just "template and go" for this
    solution. Having a data bag to stuff in options/multioptions and render to
    a template is a far nicer solution.

    THe underlying data-provider is backed by a SQLITE database on each unit,
    tracking the dictionary, provided from the 'charmhelpers' python package.

    Summary:
    opts = DockerOpts()
    opts.add('bip', '192.168.22.2')
    opts.to_s()
    '''

    def __init__(self, opts_path=None):
        self.db = unitdata.kv()
        if not self.db.get('docker_opts'):
            self.data = {}
        else:
            self.data = self.db.get('docker_opts')

    def __save(self):
        self.db.set('docker_opts', self.data)

    def add(self, key, value, strict=False):
        '''
        Adds data to the map of values for the DockerOpts file.
        Supports single values, or "multiopt variables". If you
        have a flag only option, like --tlsverify, set the value
        to None. To preserve the exact value, pass strict

        eg:
        opts.add('label', 'foo')
        opts.add('label', 'foo, bar, baz')
        opts.add('flagonly', None)
        opts.add('cluster-store', 'consul://a:4001,b:4001,c:4001/swarm', strict=True)
        '''
        if strict:
            self.data['{}-strict'.format(key)] = value

        if value:
            values = [x.strip() for x in value.split(',')]
            # handle updates
            if key in self.data and self.data[key] is not None:
                item_data = self.data[key]
                for c in values:
                    c = c.strip()
                    if c not in item_data:
                        item_data.append(c)
                self.data[key] = item_data
            else:
                # handle new
                self.data[key] = values
        else:
            # handle flagonly
            self.data[key] = None
        self.__save()

    def remove(self, key, value):
        '''
        Remove a flag value from the DockerOpts manager
        Assuming the data is currently {'foo': ['bar', 'baz']}

        d.remove('foo', 'bar')
        > {'foo': ['baz']}

        :params key:
        :params value:
        '''
        self.data[key].remove(value)
        self.__save()

    def to_s(self):
        '''
        Render the flags to a single string, prepared for the Docker
        Defaults file. Typically in /etc/default/docker

        d.to_s()
        > "--foo=bar --foo=baz"
        '''
        flags = []
        for key in self.data:
            if self.data[key] == None:
                # handle flagonly
                flags.append("--{}".format(key))
            elif '-strict' in key:
                # handle strict values
                proper_key = key.rstrip('-strict')
                flags.append("--{}={}".format(proper_key, self.data[key]))
            else:
                # handle multiopt and typical flags
                for item in self.data[key]:
                    flags.append("--{}={}".format(key, item))
        return ' '.join(flags)



from .runner import run
from .workspace import Workspace


class Compose:
    def __init__(self, workspace, strict=True):
        '''
        Object to manage working with Docker-Compose on the CLI. exposes
        a natural language for performing common tasks with docker in
        juju charms.

        :param workspace:  Define the CWD for docker-compose execution

        :param strict: - Enable/disable workspace validation
        '''
        self.workspace = Workspace(workspace)
        if strict:
            self.workspace.validate()

    def build(self, service=None, force_rm=True, no_cache=False, pull=False):
        '''
        Build or rebuild services.

        Services are built once and then tagged as `project_service`. If you
        change a service's Dockerfile or the contents of its build directory
        you can invoke this method to rebuild it.

        :param service: if provided will rebuild scoped to that service
        :param force_rm: Always remove intermediate containers.
        :param no_cache: Do not use cache when building the image
        :param pull: Always attempt to pull a newer version of the image
        '''
        cmd = "docker-compose build"

        if force_rm:
            cmd = "{} --force-rm".format(cmd)
        if no_cache:
            cmd = "{} --no-cache".format(cmd)
        if pull:
            cmd = "{} --pull".format(cmd)
        if service:
            cmd = "{} {}".format(cmd, service)

        run(cmd, self.workspace)

    def kill(self, service=None):
        '''
        Convenience method that wraps `docker-compose kill`

        :param service: if defined will only kill that service.
        '''
        if service:
            cmd = "docker-compose kill {}".format(service)
        else:
            cmd = "docker-compose kill"
        run(cmd, self.workspace)

    def pull(self, service=None):
        '''
        Pulls service images

        :param service: if defined, only pulls the image for specified service.
        '''
        if service:
            cmd = "docker-compose pull {}".format(service)
        else:
            cmd = "docker-compose pull"
        run(cmd, self.workspace)

    def restart(self, service=None):
        '''
        Restart services

        :param service: if defined, only restarts the specified service.
        '''
        if service:
            cmd = "docker-compose restart {}".format(service)
        else:
            cmd = "docker-compose restart"
        run(cmd, self.workspace)

    def rm(self, service=None):
        '''
        Convenience method that wraps `docker-compose rm`

        :param service: if defined only the specified service.
        '''
        if service:
            cmd = "docker-compose rm -f {}".format(service)
        else:
            cmd = "docker-compose rm -f"
        run(cmd, self.workspace)

    def scale(self, service, count):
        '''
        Set number of containers to run for a service.

        :param service: Service to scale as defined in docker-compose.yml
        :param count: number of containers to scale
        '''
        cmd = "docker-compose scale {}={}".format(service, count)
        run(cmd, self.workspace)

    def start(self, service):
        '''
        Start existing containers

        :param service: Service to start
        '''
        cmd = "docker-compose start {}".format(service)
        run(cmd, self.workspace)

    def stop(self, service, timeout=10):
        '''
        Stop running containers without removing them.

        :param service: Service to stop.
        :param timeout: specify a shutdown timeout in seconds.
        '''
        cmd = "docker-compose stop -t {} {}".format(timeout, service)
        run(cmd, self.workspace)

    def up(self, service=None):
        '''
        Convenience method that wraps `docker-compose up`

        :param service: if defined only launches the specified service
        '''
        if service:
            cmd = "docker-compose up -d {}".format(service)
        else:
            cmd = "docker-compose up -d"
        run(cmd, self.workspace)



from setuptools import setup

setup(
    name = "charms.docker",
    version = '0.1.6',
    author = "Charles Butler",
    author_email = "charles.butler@ubuntu.com",
    url = "http://github.com/juju-solutions/charms.docker",
    description = ( "Python wrappers for the docker CLI and configuring the Docker Daemon in Juju Charms" ),
    license = "GPLv3",
    keywords = "docker juju charm charms",
    packages = ['charms.docker'],
    long_description = "",
    classifiers = [
        "Development Status :: 3 - Alpha",
    ],
)
